from os import remove
from pathlib import Path
from typing import List, Any, Dict, Iterator, Set
from polar_vortex.interfaces.log_interface import logger
from polar_vortex.protocols.database_protocols import (
                                            DatabasePtr, 
                                            DataOperationPtr, 
                                            DataPoint, Data,
                                            Fact)
import shelve

db_path = Path(__file__).parent / 'vortex_databases'
# db_path = Path().cwd() / 'database/vortex_databases'
if not db_path.exists(): db_path.mkdir()


class VortexInterface():
    
    '''
    VortexInterface provides a uniform interface to a shelve database.
    '''
    
    def __init__(self, db_ptr:DatabasePtr,data_point:DataPoint={}) -> None:
        self.values:Dict[Fact or str, DatabasePtr] = {}
        self._del_keys:Set[Fact or str] = set()
        self.db_ptr = db_ptr
        self.values = {}
        match(data_point):
            case None:
                self.values = self.get()
            case _:
                self.upsert(data_point)
        # Shelve can create mutiple files with dame stem
        if not list(self.db_files):self.save()

    def __bool__(self,) -> bool:
        match(self.db_ptr):
            case DatabasePtr(_, None):
                return bool(self())
            case DatabasePtr(_, key):
                return bool(self()[key]) if key in self() else False
    
    def is_empty(self,) -> bool:
        return not self

    @property
    def db_files(self,) -> Iterator[Path]:
        return filter(lambda f: f.stem == self.db_ptr.database, db_path.iterdir())
    
    def detect_change(self,) -> bool:
        return True

    def verify_datum(datum:DataPoint) -> bool:
        '''
        Not implemented, return True for functionality
        '''
        return True

    def save(self,) -> bool:
        with shelve.open(str(db_path/self.db_ptr.database)) as db:
            for key in self._del_keys: del db[key]
            for key,value in self.values.items():db[key] = value
            self._del_keys = set()
        return True
    
    def upsert(self, data: Data or  DataPoint) -> bool:
        logger.debug(f'{self.db_ptr=}')
        match(self.db_ptr):
            case DatabasePtr(_, None):
                if not isinstance(data,dict): raise TypeError('data must be a dict')
                self().update(data)
            case DatabasePtr(_, key): self()[key] = data
            case _: raise Exception(f'{self.db_ptr=} is not a valid DatabasePtr')
        return True
    
    def get(self,) -> DataPoint:
        if not list(self.db_files): self.save()
        with shelve.open(str(db_path/self.db_ptr.database)) as db:
            
            match(self.db_ptr):
                case DatabasePtr(_, None): self().update(dict(db)) 
                case DatabasePtr(_, key): self.values[key] = db[key] if key in db else {}
                case _: raise Exception(f'{self.db_ptr=} is not a valid DatabasePtr')
        return self()
            
    def all(self,) -> List[Dict[str,Any]]:
        return self.get()

    def delete(self,locked:bool=True) -> bool:
        if locked: return False
        self._del_keys.union(set(self().keys()))
        self.values = {}
        return True

    def keys(self,) -> Set[str]:
        return self().keys()
     
    def __contains__(self,key:str) -> bool:
        return key in self.keys()
    
    def contains(self,key:str ) -> bool:
        return key in self
                
    def is_in(self,data_point:DataPoint) -> bool:
        return any(value == data_point for value in self().values())
    
    def __call__(self,) -> Any:
        return self.values


from .log_interface import logger
from ..protocols.database_protocols import  (
                                                DatabasePtr,
                                                DataOperationPtr,
                                                Data,Datum
                                             )
from pathlib import Path
from shutil import rmtree
from os import remove
from typing import List, Tuple, Any 
from polars import (DataFrame, 
                    LazyFrame,
                    scan_parquet, 
                    scan_csv,
                    concat,
                    col, 
                    Expr,
                    )
from dataclasses import dataclass, field
from queue import PriorityQueue
# db_path = Path('interfaces/polars_databases')
db_path = Path(__file__).parent / 'polars_databases'
if not db_path.exists(): db_path.mkdir()
logger.set_log_level('debug')
class PolarsInterface():
    '''
    This is a lazy PolarsInterface to a database built off of parquet files.
    It is an implementation of the DatabaseProtocol.
    example usage:
    '''

    def __init__(self,db_ptr:DatabasePtr,file_type:str='csv'):
        database, key = self.db_ptr = db_ptr
        if not (db_file:= db_path/database).exists():
            db_file.mkdir()
        self.key_file:Path = db_file/f'{key}.{file_type}'
        self.lazyframe:LazyFrame = LazyFrame() if self.is_empty else self.scan
        self.values:Data[Datum] = list()

    def __call__(self,) -> LazyFrame:
        return self.lazyframe

    def __bool__(self)->bool:
        return self.key_file.exists()

    @property
    def file_type(self):
        return self.key_file.suffix.lstrip('.')

    @property
    def scan(self) -> LazyFrame:
        match(self.file_type):
            case 'parquet': return scan_parquet(self.key_file)
            case 'csv': return scan_csv(self.key_file)
            case ns: raise Exception(f'{ns} is not supported for scan')
    
    @property 
    def dataframe(self)->DataFrame:
        return self().collect()
   
    @property 
    def is_empty(self)->bool:
        return not self
   
    def detect_change(self) -> bool:
        nl = '\n'
        return self.describe_optimized_plan() == \
                f'  CSV SCAN {self.key_file.name}{nl}  PROJECT */{len(self.columns)} COLUMNS{nl}'

    def save(self,has_header:bool=False)->LazyFrame:
        
        """
            Save the database to disk.
        Returns:
            LazyFrame: returns a new lazyframe pointing to the updated file.
        """
        df = DataFrame(self.values)        
        match(self.file_type):
            case 'parquet':
                self.dataframe.vstack(df).write_parquet(self.key_file)
            case 'csv' if self.is_empty or not self.detect_change:
                with open(self.key_file,mode='ab') as f:
                    df.write_csv(f,has_header=has_header)
            case 'csv':self.dataframe.vstack(df).write_csv(self.key_file)
            case ns: raise Exception(f'{ns} is not supported for scan')
        self.lazyframe = self.scan
        return self()

    def verify_datum(self,datum:Datum) -> bool:
        """
        verifies data before adding
        Args:
            datum (Datum): the piece of data. Just checks column names match ATM.

        Returns:
            bool: if the columns match
        """
        logger.debug(f'{self().columns=}')
        return datum._fields == tuple(self().columns)
    
    
    def upsert(self, 
               values:Data or Datum, 
               op_ptrs:DataOperationPtr = None) -> bool:
        """
        Upserts values into op_ptr location 
        Args:
            op_ptr (DataOperationPtrs): where the data is being upserted
            values (Data): the data being uploaded.

        Raises:
            Exception: _description_

        Returns:
            bool: _description_
        """        
        if not isinstance(values, Data):values = [values]
        datum: Datum = values[0]
        logger.debug(f'{values=}, {self.key_file}')
        match(op_ptrs):
            case _ if self.is_empty:
                self.values += values
                self.save(has_header=True)
            case _ if not self.verify_datum(datum):
                raise TypeError(f'{datum._fields} mismatch{self().columns}')                            
            case None: self.values += values
            case DataOperationPtr(indexs,None):
                lf:LazyFrame = self()
                indexs = indexs.queue
                dindexs = [di_1 - di_0 for di_0, di_1 in zip(indexs[:-1],indexs[1:])]
                lazy_ops:List[Tuple[LazyFrame]] = [(lf.slice(0,indexs[0]), DataFrame(values[0]))]
                lazy_ops += [(lf.slice(i+1,di),DataFrame(v).lazy()) 
                            for di,i,v in zip(dindexs,indexs,values)]
                
                cat_lf: List[LazyFrame] = [lf.slice(indexs[-1])]
                for op in lazy_ops:
                    cat_lf += list(op)
                cat_lf = cat_lf[1:] + cat_lf[0]
                self.lazyframe = concat(cat_lf)
                
            case DataOperationPtr(indexs, columns):
                logger.warning('''Column base operations have not been implemented. 
                                  Please dont define columns''')
                return False
        return True
            
    def get(self, op_ptr:DataOperationPtr,indexed:bool=False) -> LazyFrame:
        
        """
            Get values from the database. If a database and 
            key are provided, the entire table is returned. If a database, key,
            and index are provided, only the value at that index is returned.
        Returns:
            returns the lazyframe of the query
        """        
        
        get = self.all().with_row_count('index') if indexed else self.all()
        match(op_ptr):
            case DataOperationPtr(_, _, None): return get
            case DataOperationPtr( _,_, index):
                get = get.with_row_count('row')\
                         .filter(col('row') == index)\
                         .drop('row')
                return get
    
    def delete(self,ptr:DataOperationPtr,locked:bool=True,) -> bool:
        
        """
            Delete a value from the database. If only a database is provided,
            then whole folder is deleted. If a database and key are provided,
            then the entire table is deleted. If a database, key, and index
            then the value at that index is deleted.
        Returns:
            bool: True if successful
        """        
        
        if locked:
            logger.info('delete is locked by default')
            return False
        match(ptr):
            case DataOperationPtr(database, None, None): rmtree(db_path/database)
            case DataOperationPtr(database, key, None): remove(db_path/database/key)
            case DataOperationPtr( database,key, index) :
                lf = lf.filter(col('row') != index)
                self.lazyframe = lf.drop('row')        
        return True
    
    def all(self,ptr:DataOperationPtr=None) -> LazyFrame :
        return self.lazyframe

    def contains(self, column:str='', ptr:DataOperationPtr=None) -> bool:
        match(column,ptr):
            case column, None: return column in self.lazyframe.columns()
            case DataOperationPtr(database,None,None), '':
                return (db_path/database).exists()
            case DataOperationPtr(database,key,None), '':
                return (db_path/database).exists()
    
    def is_in(self, value:Tuple[Any],) -> bool:
        return value in set(self.all().collect().to_dict().values())



from typing import Callable
from inspect import getframeinfo, stack
from pathlib import Path
from pysimplelog import Logger
from pprint import PrettyPrinter

log_dir = Path(__file__).parent / 'logs'

def debuginfo(log_fn:Callable)->Callable:
    def wrapper(msg,*args,**kwargs):
        frameinfo = getframeinfo(stack()[1][0])
        msg = '\n'.join(map(lambda _: _.strip(),
                            filter(bool ,f'''
                                            File: {frameinfo.filename} Line: {frameinfo.lineno} Function: {frameinfo.function}
                                            {msg}
                                            {'#'*80} 
                                        '''.split('\n')
                            )))
        return log_fn(msg,*args,**kwargs)
    return wrapper    
    

logger = Logger('polar_vortex')

logger.debug = debuginfo(logger.debug)

pp = PrettyPrinter(indent=4)


logger.set_log_file_basename((log_dir/'main_logs').as_posix())

def set_log_level(level:str) -> str:
    logger.set_minimum_level(logger.logLevels[level])
    logger.debug(f'Set log level to {level}')
    
logger.set_log_level = set_log_level


from typing import NamedTuple, Any, Tuple, Dict, Set, List, Iterable
from dataclasses import dataclass, field, fields, astuple, asdict
from pathlib import Path
from interfaces.log_interface import logger
from protocols.database_protocols import DatabaseProtocol, DatabasePtr
from interfaces.polar_interface import (PolarsInterface, 
                                        LazyFrame,
                                        col,
                                        )

# root = Path(__file__).parent.parent
root = Path()
databases = root / 'databases'

class PolarModel():
    '''
    This is the base model for all models. 
    Models allow easy mapping of data to the underlying storage(ORM).
    These models are mapped to polars dataframes,stored parquet files. 
    The PolarsInterface is a lazy interface to the parquet files,
    which controls the data view.
    '''
    connection:DatabasePtr = DatabasePtr(database='default', key='default')
    _dbi = PolarsInterface
    dbi:DatabaseProtocol = _dbi(connection)

    @classmethod    
    def connect(cls, connection:DatabasePtr=None) -> bool:
        connection = cls.connection = connection or cls.connection
        cls.dbi = cls._dbi(connection)
        return True
    
    def __iter__(self):
        return iter(astuple(self))
    
    @property
    def as_tuple(self)->Tuple:
        return astuple(self)
    
    @property
    def as_dict(self)->Dict:
        return asdict(self)
    
    @property
    def as_dataframe(self)->LazyFrame:
        return self.all() 
     
    @property
    def fields(self)->Tuple:
        return map(lambda x: Field(self.connection.database,
                                   self.connection.key, 
                                   x.name, str(x.type)), 
                    fields(self))
    
    def upsert(self) -> bool:
        return self.dbi.upsert(value = self.as_dict)

    def register_model(self) -> bool:
        for field in self.fields: field.upsert()
        Field.save()
   
    def delete(self,) -> bool:
        return self.dbi.delete(DatabasePtr(value=self.as_dict))

    @classmethod
    def get(cls, db_connection:DatabasePtr,
                limit:int=2,
                indexed:bool=False,
                page_number:int=0,
                )-> Iterable[Any] or Tuple[int,Any]:
        page = cls.dbi.get(db_connection,indexed=indexed)\
                        .with_row_count()\
                        .filter(col('row_nr') >= (page_number * limit))\
                        .limit(limit)\
                        .drop('row_nr')\
                        .collect()
        while not page.is_empty():
            for row in page.iterrows():
                yield (row[0], cls(*row[1:])) if indexed else cls(*row)
            page_number += 1
            page = cls.dbi.get(db_connection,indexed=indexed)\
                            .with_row_count()\
                                .filter(col('row_nr') > (page_number * limit))\
                                .limit(limit)\
                                .drop('row_nr')\
                                .collect()
    
    @classmethod
    def all(cls,
            limit:int=2,
            indexed:bool=False,
            page_number:int=0,
            ) -> List[Tuple]:
        database, key, _, _ = cls.connection
        return cls.get(DatabasePtr(database,key,None,None),limit,indexed,page_number)

    @classmethod
    def contains(cls, key:str='')->bool:
        key = key or cls.key
        return cls.dbi.contains(key) 
    
    @classmethod
    def is_in(cls, value:Tuple[Any], key:str='')->bool:
        key = key or cls.key
        return cls.dbi.is_in(cls.database,value,key) 
    
    @classmethod
    def save(cls) -> bool:
        return cls.dbi.save()

@dataclass(frozen=True,slots=True)
class Field(PolarModel):
    database:str
    key:str
    name:str
    Type: str
    
    class Meta(PolarModel.Meta):
        connection = DatabasePtr('model','fields')
        
Field.Meta.connect()


from typing import (
                    Protocol,
                    Any,Set,
                    Optional,
                    NamedTuple,
                    List, Dict,
                    TypedDict,
                    TypeAlias
                    )
from queue import PriorityQueue

Fact: TypeAlias = NamedTuple
Datum: TypeAlias = TypedDict
DataPoint: TypeAlias = Fact or Datum or Dict or int or float or str or bool or None
Data: TypeAlias = List or List[DataPoint]
class DatabasePtr(Fact):
    '''
    DatabaseConnections provides a uniform method to pass input to all the functions of the protocol. 
    The design pattern is Protocol defines actions, the Interface provides a uniform function interface.
    '''
    database:Optional[str] = None
    key: Optional[str] = None

class DataOperationPtr(Fact):
    indexs:Optional[PriorityQueue[int] or int] = None
    columns:Optional[Set[str] or DataPoint] = None

class DatabaseProtocol(Protocol):

    '''
    DatabaseProtocol defines the actions that can be performed on a database.
    '''
    
    db_ptr: DatabasePtr
    values: Data or DataPoint
    
    def detect_change(self,) -> bool:
        """
        detect_change in underlying database. Useful for reducing reads.
        Returns:
            bool: if the underlying database changed,
                    return True
        """        
        ...
        
    def save(self,)-> Any:
        '''
        Commits a database to disk. This could a bulk save.
        '''
        ...
    
    def verify_datum(datum:DataPoint) -> bool:
        """
        verifies data before adding
        Args:
            datum (Datum): the piece of data. Just checks column names match ATM.

        Returns:
            bool: if the columns match
        """ 
        ...    

    def upsert(self,
            data: Data or DataPoint,
            op_ptrs:Optional[DataOperationPtr]
            ) -> bool:
        '''
        set a value in the database with a key and possible index
        '''
        ...

    def get(self,
            op_ptrs:Optional[DataOperationPtr]
            ) -> Any:
        '''
        get a value from the database with a key and possible index
        '''
        ...

    def delete(self, 
               op_ptrs:Optional[DataOperationPtr],
               locked:bool=True,
               ) -> bool:
        '''
        delete a value from the database with a key and possible index
        '''
        ...
        
    def all(self, ) -> Any:
        '''
        Get all values from the database with a key
        '''
        ...
   
    def contains(self,
                op_ptrs:Optional[DataOperationPtr] or Fact or str,
                 )-> bool:
        '''
        Checks if databases exist or if a key is a database or if column is in key
        '''
        ... 
       
        
    def is_in(self,
              data:DataPoint,
                 )-> bool:
        '''
        checks if a value is in db object
        '''
        ...
        
    def is_empty(self,)-> bool:
        '''
        checks if a value is in db object
        '''
        ...